////+------------------------------------------------------------------+
////| Generic comparison function Less.                                |
////+------------------------------------------------------------------+
//template<typename T>
//bool Less(const T left,const T right)
//  { return left < right;  }


//+------------------------------------------------------------------+
//| Optimized Merge Sort with insertion (Hybrid)                     |
//+------------------------------------------------------------------+
template<typename T>
void MergeSort(T &arr[])
{
    T temp[];
    // Create a copy of the original array. Switching between
    // original array and its copy will allow to avoid
    // additional array allocations and data copying.
    ArrayCopy(temp, arr);
    MergeSortInternal::SplitMergeInsertion(arr, temp, 0, ArraySize(arr) - 1);
}



namespace MergeSortInternal
{
//+------------------------------------------------------------------+
template<typename T>
void SplitMergeInsertion(T &arr[], T &temp[], int lo, int hi)
{
    const int InsertsortSizeThreshold = 16;

    // If subarray size is low then do insertion sort
    if (hi - lo + 1 <= InsertsortSizeThreshold)
    {
        InsertionSort(arr, lo, hi);
        return;
    }
    int mid = lo + (hi - lo) / 2;
    SplitMergeInsertion(temp, arr, lo, mid);
    SplitMergeInsertion(temp, arr, mid + 1, hi);
    Merge(arr, temp, lo, mid, hi);
}
//+------------------------------------------------------------------+
// Merge two sorted sub arrays src[lo..mid] and src[mid+1..hi] to dst[lo..hi]
template<typename T>
void Merge(T &dst_array[], T &src_array[], int lo, int mid, int hi)
{
    // Stop if already sorted.
    // Is biggest element in first half <= smallest element in second half?
    // Helps for nearly ordered lists (more adaptive).
    if (mid + 1 < hi && !Less(src_array[mid + 1], src_array[mid]))
    {
        ArrayCopy(dst_array, src_array, lo, lo, hi - lo + 1);
        return;
    }
    int i = lo, j = mid + 1;
    for (int k = lo; k <= hi; k++)
    {
        if (i > mid)
        {
            dst_array[k] = src_array[j++];
        }
        else if (j > hi)
        {
            dst_array[k] = src_array[i++];
        }
        else
        {
            dst_array[k] = (Less(src_array[i], src_array[j]) ? src_array[i++] : src_array[j++]);
        }
    }
}
//+------------------------------------------------------------------+
template<typename T>
void InsertionSort(T &arr[], int lo, int hi)
{
    for (int i = lo + 1; i <= hi; i++)
    {
        // Compare first so we can avoid 2 moves for an element already positioned correctly.
        // More adaptive for already sorted data - i.e. never assign an element to itself!
        if (Less(arr[i], arr[i - 1]))
        {
            T curr = arr[i];
            arr[i] = arr[i - 1];
            int j = i;

            while (--j > lo && Less(curr, arr[j - 1]))
                arr[j] = arr[j - 1];

            arr[j] = curr;
        }
    }
}
//+------------------------------------------------------------------+
}
